#
#   Smart Gateway Database validation
#
# Depends on dbfunctions and updateDB loaded current


import time
from dbfunctions import write_control_t,dbsql2,dbsqlw,add_column_if_missing
from datetime import date, datetime
from updateDb import updateDb
from utilities import printm, set_prefix, get_prefix

# Open DB Connector
# dbchecker Version
Version = 1.53

def dbcheck(cnx):
    #
    #
    #   Check integrity of fifo table
    #
    calling_prefix =  get_prefix()
    set_prefix('DBCheck')
    printm( ' Running FIFO Integrity Test.')
    row = dbsql2(0,'check table fifo medium;')
   # print row

    #cnx = dbopen()
    cnx = 0
    if row[3] == 'OK':
        printm(' FiFO table checks OK.')
    else:
        # Operation Failed
        printm(' Fifo table failed check!')
        cursor = dbsqlw(cnx,'drop table if exists fifo')
        cursor = dbsqlw(cnx,"CREATE TABLE `monitor`.`fifo` ( `idfifo` INT NOT NULL AUTO_INCREMENT,  `Sent`  INT NULL,`CTime` DATETIME NULL,  `CData` VARCHAR(4000) NULL, `Smqtt` INT NULL,  PRIMARY KEY (`idfifo`));")
        #row = cursor.fetchone()

    #
    #   Check integrity of control table
    #
    row = dbsql2(0,'check table control medium;')
    printm( row)

    #cnx = dbopen()
    cnx = 0
    if row[3] == 'OK':
        printm( ' Control table checks OK.')
    else:
        printm( ' Control table failed check!')
        cursor = dbsqlw(cnx,'drop table if exists control')
        cursor = dbsqlw(cnx,"CREATE TABLE `monitor`.`control` (`idcontrol` INT NOT NULL AUTO_INCREMENT,`Control` VARCHAR(45) NULL,`Cnumber` INT NULL,`Ctext` VARCHAR(400) NULL, PRIMARY KEY (`idcontrol`));")
#
    #   Check integrity of sensorlist table
    #
    row = dbsql2(0,'check table sensorlist medium;')
    printm( row)

    #cnx = dbopen()
    cnx = 0
    if row[3] == 'OK':
        printm( ' sensorlist table checks OK.')
        add_column_if_missing("sensorlist","RxTime","Datetime")
    else:
        printm( ' sensorlist table failed check!')
        cursor = dbsqlw(cnx,'drop table if exists sensorlist')
        cursor = dbsqlw(cnx,"CREATE TABLE `monitor`.`sensorlist` (`id` INT NOT NULL AUTO_INCREMENT,`deviceid` VARCHAR(10) NULL,`mqttkey` VARCHAR(15) NULL,`lvalue` VARCHAR(10) NULL,`count` INT NULL,`RxTime` DATETIME NULL, PRIMARY KEY (`id`));")

    #   Check integrity of Sensorerrors table
    #
    row = dbsql2(0,'check table Sensorerrors medium;')
    printm( row)

    #cnx = dbopen()#

    cnx = 0
    if row[3] == 'OK':
        printm( ' Sensorerrors table checks OK.')
    else:
        printm( ' Sensorerrors table failed check!')
        cursor = dbsqlw(cnx,'drop table if exists Sensorerrors')
        cursor = dbsqlw(cnx,"CREATE TABLE `monitor`.`Sensorerrors` (`id` INT NOT NULL AUTO_INCREMENT,`deviceid` VARCHAR(10) NULL,`issue` VARCHAR(25) NULL,`count` INT NULL,`RxTime` DATETIME NULL, PRIMARY KEY (`id`));")

    #
    #  Check FIFO Indexes Exists
    #
    sql="show keys from fifo where Key_name='iSent'"
    row = dbsql2(0,sql)
    if row is None:
        printm( ' Creating index: iSent')
        sql ="create index iSent on fifo(Sent)"
        cursor = dbsqlw(cnx,sql)
    else:
        printm( ' Index iSent in fifo')

    sql="show keys from fifo where Key_name='iCtime'"
    row = dbsql2(0,sql)
    if row is None:
        printm( ' Creating index: iCtime')
        sql ="create index iCtime on fifo(Ctime)"
        cursor = dbsql2(cnx,sql)
    else:
        printm( ' Index iCtime in fifo')
    #
    #  Verify all Control Keys
    updateDb(cnx)   # make sure we have all db fields defined


    # Store version and run time of db check
    #
    write_control_t(cnx,'dbcVersion',Version)
    dstart = time
    Start =dstart.strftime("%m/%d/%Y %H:%M:%S")
    write_control_t(cnx,'dbcLastRun',Start)

    printm( ' dB Health Check completed')

    # Restore prefix
    set_prefix(calling_prefix)
