#! /usr/bin/python

# Commander checks in to Monitoring server to report alive and working
# Sends dump of all controls for remote monitoring
# receives commands from server
# Monitors WiFi and restores connection if lost

#Upgrade to current OS and Latest Python and MySQL

__author__ = "glenn"
__date__ = "$Mri 16, 2024 "
import os
import linecache
import sys
import time
from time import mktime
import json
import requests
#import MySQLdb
import socket
import unicodedata

from updateDb import updateDb
from dbfunctions import * #dbopen,write_control_n,write_control_t,write_control,read_control,dbsql2,dbsqlw,dbopeninit
from utilities import * #getgwid, set_prefix, printm
from datetime import date, datetime

set_prefix('CMD')
CmdVersion = '2.51'
rxsend = 0
cnx = dbopeninit()
#updateDb(cnx)   # make sure we have all db fields defined
printm(CmdVersion, ' Started')

def updateip():
    gw = os.popen("ip -4 route show default").read().split()
    printm('Show route: ',len(gw))
    if len(gw) > 3:
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.connect((gw[2], 0))
        ipaddr = s.getsockname()[0]
        gateway = gw[2]
        host = socket.gethostname()
    else:
        ipaddr = '0.0.0.0'

    printm('System IP Address is: '+str(ipaddr))
    # set current IP in table
    write_control_t(0,'IP',ipaddr)
    return str(ipaddr)

def remove_control_characters(line):
    return ''.join(c for c in line if ord(c) >= 32)
#
#   Gateway Command Functions
#
def SendStatus():
    global url, tagt, senderror,senderrordelay , senderrorcount
    # Capture latest Temperatures of GPU and CPU
    gputemp = os.popen('vcgencmd measure_temp').readline()
    gputemp = gputemp[5:9]
    gputempc = float(gputemp)
    gputemp =  gputempc * 9.0 / 5.0 + 32.0
    gputemp = round(gputemp,1)
    cputemp = os.popen('cat /sys/class/thermal/thermal_zone0/temp').readline()
    cputempc = float(cputemp)/1000
    cputemp =  cputempc * 9.0 / 5.0 + 32.0
    cputemp = round(cputemp,1)
    #cnx = dbopen()
    cnx=0
    write_control_t(cnx,'GPUTemp',gputemp)
    write_control_t(cnx,'CPUTemp',cputemp)
    #
    # No Server then Don't send
    if url.find("http") == -1:
        return 0
    # make sure what IP we are using
    if updateip() == '0.0.0.0':
        return 0
    #
    # update status with current fifo count
    printm( 'Get fifo count')
    #cnx = dbopen()
    row = dbsql2(0,'select count(idfifo) from fifo')
    FIFOCount = row[0]
    printm( 'Fifo count is: ',FIFOCount)
    write_control_n(cnx,'FIFOCount',FIFOCount)
    row = dbsql2(0,'select count(idfifo) from fifo where sent = 0')
    FIFOQueue = row[0]
    write_control_t(cnx,'FIFOQueue',FIFOQueue)
    #cnx = dbopen()
    printm("Set Transaction")
    dbsqlw(cnx,'SET TRANSACTION ISOLATION LEVEL READ UNCOMMITTED;')
    params = '{ "Gateway": "'+mac[:17]+'","Time":"'+tagt
    rows = dbsql2(1,'select * from control')
    for row in rows:
        params = params + '","'+row[1]+'":"'+str(row[2])+' - '+remove_control_characters(row[3])

    params = params + '"}'
#    print 'cmd: ',params

    params = dict(data=params)

    if senderror == 0:
        printm(params)
        try:
            rxsend = 1
            resp = requests.post(url=url, data=params)
            rxsend = 0

            if r.status_code == 200:
        #    print resp.url
                data = json.loads(resp.text)
            else:
                printm( '***Missing Reply for Status Message ***')

        except requests.exceptions.Timeout:
            if senderror == 0:
                print("CMD Status Update timed out")
                CheckChome()
            senderror += 1
            pass
        except:
            if senderror == 0:
                printm( 'Send Status Message Failed')
                CheckChome()
            senderror += 1
            pass

        return 0
    else:
        printm('Send Checkin Errors, delay sending')
#
def ShowSSID():
    from subprocess import check_output
    global tagt

    params = '{ "Gateway": "'+mac[:17]+'","Time":"'+tagt+'","SSIDs":['
    scanoutput = check_output(["iwlist", "wlan0", "scan"])
    ssid=[0]
    firstssid=1
    channel=''
    for line in scanoutput.split():
        line = line.decode("utf-8")
        if line[:7] == "Channel":
            channel = line.split(':')[1]
        if line[:5]  == "ESSID":
            onessid = line.split('"')[1]
            ssid.append(onessid)
            if firstssid == 0:
                params = params +','
            params = params + '{ "SSID": "'+onessid+'","Channel": "'+channel+'"}'
            firstssid=0
            channel=''
    params = params + ']}'
    params = dict(data=params)
    printm(params)

    rxsend = 1
    resp = requests.post(url=url, data=params)
    rxsend = 0
#    print resp.url
    if r.status_code == 200:
#    print resp.url
        data = json.loads(resp.text)
    else:
        printm('***Missing Reply for SSID Message ***')

#
# Show WiFi Settings
def ShowWiFi():
    global tagt
    row = read_control(0,'WiFi')
    WiFi = row[3]
    row = read_control(0,'Psk')
    Psk = row[3]

    params = '{ "Gateway": "'+mac[:17]+'","Time":"'+tagt+'","WiFi":"'+str(WiFi)+'","Psk":"'+str(Psk)+'"}'
    params = dict(data=params)
    printm(params)

    rxsend = 1
    resp = requests.post(url=url, data=params)
    rxsend = 0
#    print resp.url
    if r.status_code == 200:
        data = json.loads(resp.text)
    else:
        data ='No Reply'
        printm( 'Error sending Status')

#
def SetWiFi():
 #
    '''
    target = open("wpa_supplicant.base",'r')
    source = target.read()
    printm(source)

    target.close()
    target = open("wpa_supplicant.conf",'w')
    target.write(source)

    row = read_control(cnx,'WiFi')
    WiFi = row[3]
    row = read_control(cnx,'Psk')
    Psk = row[3]

    newWiFi = '  ssid="'+str(WiFi)+'"\n'
    newWiFi = newWiFi + '  psk="'+str(Psk)+'"\n'
    newWiFi = newWiFi + '}\n'
    target.write(newWiFi)
    target.close()
    #
    # Be sure put in WIFI Mode and set WIFI SSID
    os.system("sudo /home/pi/gw2/wifirestore.sh")
    '''
    return
 #
def JoinZtier():
    # Test if Zerotier app is installed

    # Issue join
    network = Value

    # Update Control with new network

def LeaveZtier():
    # Verify Zerotier app is installed

    # Issue Leave Command
    network = Value

    # Update Control remove network


def SetUpdate():
    dbTfields =['IP','WiFi','Psk','Proto','Pair','Dhome','Chome','WRLIPList','WRLPortList','WRLAppList','WRLModelList']
    dbNfields =['Fsend','Nsend','Srate','ExecptionsCmd','ExecptionsLog','ExecptionsSend','WiFiDrops','LEDon','WRCycle','LEDLevel','FlushCycle','FlushLimit']
    if Control in dbNfields:
        printm('Update Number field: ',Control,' with ', Value)
        write_control_n(0,Control,Value)
    elif Control in dbTfields:
        printm('Update Text field: ',Control,' with ', Value)
        write_control_t(0,Control,Value)
    else:
        printm('Invalid Control Field')

#
def ResetGateway():
    os.popen("sudo reboot")

#
def HaltGateway():
    os.popen("sudo shutdown -h -t 0 now")

#
def OSRelease():
    global url, tagt
    # No Server then Don't send
    if url.find("http") == -1:
        return 0
    space = os.popen("sudo df |grep /dev |head -1")
    freespace = space.readline()
    Uos = os.popen("uname -or")
    Ulicense = Uos.readline()
    Urel = os.popen("lsb_release -irc")
    Urelease1 = Urel.readline()
    Urelease1 = Urelease1.replace('\t','')
    Urelease2 = Urel.readline()
    Urelease2 = Urelease2.replace('\t','')
    Urelease3 = Urel.readline()
    Urelease3 = Urelease3.replace('\t','')
    UResp = Ulicense[:-1] +" : "+Urelease1[:-1]+" : "+Urelease2[:-1]+" : "+Urelease3[:-1]
    gethwplatform = os.popen('cat /sys/firmware/devicetree/base/model')
    HWPlatform = gethwplatform.readline()
    HWPlatform = HWPlatform[:len(HWPlatform)-1]
    params = '{ "Gateway": "'+mac[:17]+'","Time":"'+tagt+'","DiskFree":"'+freespace[:-2]+'","OSVersion":"'+UResp+'","HWPlatform":"'+HWPlatform+'"}'
    params = dict(data=params)
    printm(params)

    try:
        rxsend = 1
        resp = requests.post(url=url, data=params)
        rxsend = 0
        if r.status_code == 200:
    #    print resp.url
            data = json.loads(resp.text)
        else:
            printm( '***Missing Reply for OSRelease Message ***')
    except:
        printm( 'Send OS Release Failed')
        pass

    return 0

def Terminate():
    executeit = 0
    cnx = 0
    if Value == 'Log':
        row = read_control(cnx,'LogPID')
        executeit = 1
    if Value == 'Send':
        row = read_control(cnx,'SendPID')
        executeit = 1
    if Value == 'HealthCk':
        row = read_control(cnx,'HCPID')
        executeit = 1
    if Value == 'Cmd':
        row = read_control(cnx,'CmdPID')
        executeit = 1
    if Value == 'term-HealthCk':
        row = read_control(cnx,'HCPID')
        executeit = 1
    if Value == 'term-gatewaywrl':
        row = read_control(cnx,'WRPID')
        executeit = 1
    #if Value == 'term-gatewaytarts':
    #    row = read_control(cnx,'TartgwPID')
    #    executeit = 1
    #if Value == 'term-loader':
     #   row = read_control(cnx,'LoaderPID')
     #   executeit = 1
    if executeit == 1:
        PID = row[2]
        printm( 'Terminating: ',Value," PID: ",PID)
        killcmd = "sudo kill "+str(PID)
        printm( 'Kill Command: ',killcmd)
        os.system(killcmd)
    else:
        printm( 'Terminate Failed with ',Value)

def SetSDate():
    global tagt
    setsystemdate = os.system("sudo date -s '"+Value+"'")
    getDate = os.popen("sudo date")
    newdater = getDate.readline()
    newdate = newdater.translate(None,'\n')
    params = '{ "Gateway": "'+mac[:17]+'","Time":"'+tagt+'","NewDate":"'+newdate+'"}'
    params = dict(data=params)
    printm( 'New Time reply: ',params)

    rxsend = 1
    resp = requests.post(url=url, data=params)
    rxsend = 0

def SetWirelessId():
    # set new wireless Id
    # limit to 2 characters
    #
    global Value,wtactive,wtactivelist
    Value = Value[:2]
    #cnx = dbopen()
    cnx = 0
    write_control_t(cnx,'SetId',Value)
    write_control_n(cnx,'ProcessId',0)
    write_control(cnx,'WTactive',wtactive,wtactivelist)

def ClearWirelessId():
    #cnx = dbopen()
    cnx = 0
    write_control_t(cnx,'SetId','')
    write_control_n(cnx,'ProcessId',1000)

def CheckChome():
    global url, senderror,senderrordelay , senderrorcount
    #cnx = dbopen()
    cnx = 0
    row = read_control(cnx,'Server')
    if len(row[3]) < 3:
        url = ""
        return
    nurl = row[3]
    row = read_control(cnx,'Chome')
    nurl = nurl + row[3]
    #print(f"url --{url}-- nurl --{nurl}--")
    if url != nurl:
        url = nurl
        printm('New Send Data to: '+url)
        senderror = 0
        senderrordelay = senderrorcount

#
def SendEmail():
    global mac
    printm( 'Execute command: ',Value)
    os.system(Value + " |mail -s 'Gateway "+mac+" Message' glenn.seaton@seltron.com &")

def TestLED():
    #
    # Shutdown Healthcheck
    return

def SetTempScale():
    #
    global Value,wtactive,wtactivelist
    printm('Set Temperature Scale to: ',Value)
    if Value == 'C' or Value == 'F':
        #cnx = dbopen()
        cnx = 0
        write_control_t(cnx,'TempScale',Value)
        if Value == 'F':
            write_control_n(cnx,'TempScale',1)
        else:
            write_control_n(cnx,'TempScale',0)
    else:
        printm('Invalid Temperature Scale Setting')

#
# Get Server Status
serverstatus = read_control(cnx,'ServerStatus')
serverstatus = serverstatus[2]

#
IPAddress = updateip()

Failcode = 0
if len(IPAddress) < 8:
    Failcode = 2
cnx = 0
write_control_n(cnx,'CmdState',Failcode)
write_control_t(cnx,'CmdVersion',CmdVersion)

Srate = 20
printm("Sample Rate set to: "+str(Srate)+" seconds")
Fsend = 10
printm("No Change Check send set to: "+str(Fsend)+" seconds")

write_control_n(cnx,'ButtonCmd',0)

mac = getgwid()
printm(mac)

#cursor.execute("select * from control where Control ='Chome'")
row = read_control(cnx,'Chome')
url = row[3]
row = read_control(cnx,'Server')
url = row[3] + url
printm('Send Data to: '+url)

#
write_control_n(cnx,'CmdPID',str(os.getpid()))
printm( 'Process Id = ',os.getpid())
#

#
#  Set initial constants
#
sendcycle = 1
sendnow = sendcycle
dstart = time
Start =dstart.strftime("%m/%d/%Y %H:%M:%S")
#cursor.execute("update control set Ctext ='"+Start+"' where Control = 'CmdStart' ")
#cnx.commit()
write_control_t(cnx,'CmdStart',Start)
ButtonCmd = 0
maxwifidrop = 8
wififailcount = maxwifidrop
TimeSendStatus = 300               # send Status every 5 minutes
LastSendStatus = time.time()-285   # send first status on second checkin

senderror = 0
senderrorcount = 2                 # How many skip 2 bulk sends (5 min each) before retry
senderrordelay = senderrorcount

nodeid = mac[:17]
nodeid = nodeid.upper()
nodeid = nodeid.replace(":","")
oldnodeid = read_control(0,'NodeId')
if oldnodeid[2] == 0:
    write_control_t(cnx,'NodeId',nodeid)
    printm('NodeId set to: '+nodeid)

printm('Commander Starting Loop')

while True:
        try:
            Gateway = 'noforme'
            Control = 'NoRx'
            Value = 'NoData'
            rxsend=0
            cnx = 0
            dstart = time
            # No Server then Don't send
            tagt= datetime.fromtimestamp(mktime(time.gmtime())).strftime('%Y-%m-%d %H:%M:%S')
            if url.find("http") != -1:
                tagm =' Id ='+mac[:17]
                tag = tagt+tagm
                msg="checkin"
            #
                jsond = '{ "Gateway": "'+mac[:17]+'","Time":"'+tagt+'","msg":"'+msg+'"}'
                #
                jsond = dict(data=jsond)

                if senderror == 0:
                    printm( jsond)
                    try:
                        rxsend =1
                        r =requests.post(url,data=jsond, timeout=5)
                        rxsend = 0
                        if r.status_code == 200:
            #                print 'cmd: ',r.text
                            wififailcount = maxwifidrop     #network working reset fail count
                            j=json.loads(r.text)
                            for item in j:
                                reply = item['Reply']
                                Gateway = item['Gateway']
                                Control = item['Control']
                                Value = item['Value']
                            if Failcode != 0:
                                Failcode = 0
                                write_control_n(cnx,'CmdState',Failcode)
                        else:
                            reply = '**Missing**'
                            Failcode = 5
                            write_control_n(cnx,'CmdState',Failcode)
                #
                #   Give it some time to recover or reset
                            time.sleep(10)
                            CheckChome()

                        printm( 'The Reply Command is: '+reply+' with Control of: '+Control+' and Value of: '+Value)

                    except requests.exceptions.Timeout:
                        if senderror == 0:
                            print("Check-in Request timed out")
                            reply = '**Failed**'
                            time.sleep(10) #Allow time for rx error to clear
                            CheckChome()
                        senderror += 1
                            # Server Online, check status
                        if serverstatus == 1:
                            serverstatus = 0
                            write_control_n(cnx,'ServerStatus',serverstatus)
                        pass
                    except:
                        if senderror == 0:
                            printm( 'Tx/Rx Failt with Check-in ')
                            reply = '**Failed**'
                            Failcode = 5
                            write_control_n(cnx,'CmdState',Failcode)
                            time.sleep(10) #Allow time for rx error to clear
                            CheckChome()
                        senderror += 1
                        # Server Online, check status
                        if serverstatus == 1:
                            serverstatus = 0
                            write_control_n(cnx,'ServerStatus',serverstatus)
                        pass
        #
        # Process command
        #
                    if Gateway == mac[:17] :
                        if reply == 'OK':
                            printm(tagt+' No Commands for now')
                            # Server Online, check status
                            if serverstatus == 0:
                                serverstatus = 1
                                write_control_n(cnx,'ServerStatus',serverstatus)
                        elif  reply == 'Status':
                            printm("send Status back to server")
                            SendStatus()
                        elif  reply == 'Update':
                            printm('Receiving Database Update')
                            SetUpdate()
                        elif  reply == 'ShowSSID':
                            printm('Send Visible WiFi SSIDs to server')
                            ShowSSID()
                        elif  reply == 'ShowWiFi':
                            printm('Send back current WiFi settings')
                            ShowWiFi()
                        elif  reply == 'SetWiFi':
                            printm('Set WiFi and enable')
                            SetWiFi()
                        elif  reply == 'ResetGateway':
                            printm('Restart Gateway')
                            ResetGateway()
                        elif  reply == 'HaltGateway':
                            printm('Halting Gateway')
                            HaltGateway()
                        elif  reply == 'OSRelease':
                            printm('OS Release and Disk Space Request')
                            OSRelease()
                        elif  reply == 'Terminate':
                            printm('Shutdown Gateway App')
                            Terminate()
                        elif  reply == 'SetSDate':
                            printm('Set System Date')
                            SetSDate()
                        elif  reply == 'SetID':
                            printm('Set New Wireless Id')
                            SetWirelessId()
                        elif  reply == 'ClearID':
                            printm('Clear New Wireless Id')
                            ClearWirelessId()
                        elif  reply == 'Send':
                            printm('Execute Command snd email results')
                            SendEmail()
                        elif  reply == 'TestLED':
                            printm('Execute Command to Test LED')
                            TestLED()
                        elif  reply == 'JoinZtier':
                            printm('Join Zerotier Network')
                            JoinZtier()
                        elif  reply == 'LeaveZtier':
                            printm('Leave Zerotier Netowrk')
                            LeaveZtier()
                        elif  reply == 'TempScale':
                            printm('Set New Temperature Scale')
                            SetTempScale()
                        else:
                            printm('Unknown Command')
                    else:
                        printm('Command for another Unit')
                else:
                    printm('Send Server Errors, delay sending')
                    if senderrordelay > 0:
                        senderrordelay -= 1
                    else:
                        senderror = 2
                        senderrordelay = senderrorcount 
            else:
                printm("Skipped, No Server")
                if serverstatus == 1:
                    serverstatus = 0
                    write_control_n(cnx,'ServerStatus',serverstatus)

    # Is it time to send Status
            if (time.time() - LastSendStatus) > TimeSendStatus/3:
                CheckChome()
                senderror = 0

            if (time.time() - LastSendStatus) > TimeSendStatus:
                SendStatus()
                if rxsend == 0:
                    OSRelease()
                if rxsend == 0 and url.find("http") != -1:
                    LastSendStatus = time.time()
                    write_control_t(cnx,'LastCmdSend',dstart.strftime("%m/%d/%Y %H:%M:%S"))
                else:
                    updateip()

            #
            #  See if Controls changed
            #
                #cnx = dbopen()
                cnx = 0
                NSrate = read_control(cnx,'Srate')
                NSrate =NSrate[2]
                if NSrate != Srate:
                    Srate = NSrate
                    printm("Sample Rate set to: "+str(NSrate)+" seconds")


#
#  Start delay till next cycle
#
            dstart = dstart.strftime("%S")
            dnow = time
            dnowsec = dnow.strftime("%S")
            dnowsec = int(dnowsec)
            if int(dstart) > dnowsec :
                dnowsec = dnowsec + 60
            while (int(dnowsec)-int(dstart)) < Srate:
                time.sleep(1)
                dnow = time
                dnowsec = dnow.strftime("%S")
                dnowsec = int(dnowsec)
                if int(dstart) > dnowsec:
                         dnowsec = dnowsec + 60

#  Trap on Control C abort, clear buffers and close database

        except KeyboardInterrupt:
                printm('Exiting')
                #cursor.close()
                cnx.close()
                sys.exit()

#
#   trap errors and log count, then wait 30 seconds and try again
#  Toggle LEDs to Red

        except Exception as inst:
            exc_type, exc_obj, tb = sys.exc_info()
            lineno = tb.tb_lineno
            errorstart = datetime.now()
            ErrorTime =errorstart.strftime("%m/%d/%Y %H:%M:%S")
            printm('At:{} EXCEPTION IN ( LINE {} ,Values {},  {} '.format(ErrorTime,lineno, inst,  type(inst)))

        # Button Command not issue, log issue
            with open("/home/pi/gw2/errorcmd.log","a") as text_file:
                text_file.write(" {}  EXCEPTION IN ({}, LINE {} ): {} \n".format(ErrorTime, filename, lineno,line.strip() ))
            printm(type(inst))
            with open("/home/pi/gw2/errorcmd.log","a") as text_file:
                text_file.write(" {} instruction {} \n".format(ErrorTime,type(inst)))
            printm(inst.args)
            with open("/home/pi/gw2/errorcmd.log","a") as text_file:
                text_file.write(" {} arguments {} \n".format(ErrorTime,inst.args))
            printm( 'WiFi Fail Count is: ',wififailcount)
            #            cursor.execute("select * from control where Control ='ExecptionsCmd'")
            row = read_control(cnx,'ExecptionsCmd')
            drops = row[2]+1
            write_control_n(cnx,'ExecptionsCmd',drops)

            Failcode = 5
            if len(IPAddress) < 8:
                Failcode = 2
            write_control_n(cnx,'CmdState',Failcode)
            #
            # Are we wired or wireless
            row = read_control(0,"WiFi")
            if len(row[3]) > 3 :
                if rxsend == 1:
                    row = read_control(cnx,'WiFiDrops')
                    drops = row[2]+1
                    write_control_n(cnx,'WiFiDrops',drops)
                    wififailcount = wififailcount-1
                    if wififailcount == 0:   #Since network has failed for awhile reset WiFi
                            # Look for USB to reset wifi
                            if os.system("iwconfig wlan0 |grep -i Frequency") > 0:
                                SetWiFi()
                            wififailcount = maxwifidrop
                            time.sleep(10)
                else:
                    #
                    time.sleep(30)
            #
            rxsend = 0
            #
            if len(IPAddress) < 8:
                IPAddress = updateip()
            write_control_n(cnx,'CmdState',0)

            pass
            raise
