 #! /usr/bin/python

# Overseer
# Monitor Status for running apps
# Power led red - green, yellow if all apps are not running
# mqtt led off -not enabled, red enabled not connected, green enabled and connected
# server led off -not enabled, red enabled not connected, green enabled and connected


__author__ = "glenn"
__date__ = "$Jun 2, 2024 PM$"

import os
import sys
import time
from time import mktime
#from typing import Sequence
import RPi.GPIO as GPIO
from datetime import date, datetime, timedelta
from dbfunctions import dbopen,write_control_n,write_control_t,write_control,read_control,dbsql,dbsql2
from utilities import  startapp, text2time, init_tmp2, set_prefix,printm
from operator import length_hint

LEDVersion = '1.71'
set_prefix("Led")

printm(LEDVersion, ' Started')

dstart = datetime.now()
Start =dstart.strftime("%m/%d/%Y %H:%M:%S")
write_control_t(0,'LEDStart',Start)
write_control_t(0,'LEDVersion',LEDVersion)

rxsend = 0
GPIO.setwarnings(False)
GPIO.setmode(GPIO.BCM)
#
# Relay pin 20 GPIO 24
GPIO.setup(24,GPIO.OUT)
GPIO.output(24,False)       #Relay On
#
#
#  Enable Led
#power  led pin 40 38 gpio 20 21
#mqtt   led pin 37 35 gpio 26 19
#server led pin 33 31 gpio 13 6
#
# power
GPIO.setup(21,GPIO.OUT)
GPIO.setup(20,GPIO.OUT)
#mqtt
GPIO.setup(26,GPIO.OUT)
GPIO.setup(19,GPIO.OUT)
#server
GPIO.setup(13,GPIO.OUT)
GPIO.setup(6,GPIO.OUT)

#
# Set Power Up to Green

GPIO.output(21,False)
GPIO.output(20,True)
#
# LED Power Status will be Green unless Wireless, Commander, or Sender are not running
# MQTT LED Will be Red if Enabled and Green if Enabled and Status On, else Off
# Server LED will be Red if not blank, Green if Connected else Off

def appstate(app_name):
    global applist
    applist = init_tmp2()
    if app_name not in applist[:]:
        status = "Down"
    else:
        status = "Up"
    return status

def ckServerStatus():
    global serverstate,serverenabled,serverfailedat
    cnx= 0
    #    print 'Mon: Log State: ',LogState
    row = read_control(cnx,'Server')
    server = row[3].strip()
    if len(server) > 5:
        serverenabled = 1
        row = read_control(cnx,'ServerStatus')      # Server Defined, is it Online
        serverstatus = row[2]
        if serverstatus == 1:
            if serverstate != 1:
                # enabled and connected show Green
                serverstate = 1
                GPIO.output(13,False)
                GPIO.output(6,True)
                printm("Turn Server Status to Green")
        else:
            # Enabled but not connected RED
            serverfailedat = time.time()
            serverstate = 0
            GPIO.output(13,True)
            GPIO.output(6,False)
            printm("Turn Server Status to Red")
    else:
        # Not enabled, turn off
        if serverstate != 0 or serverenabled == 1:
            serverstate = 0
            serverenabled = 0
            GPIO.output(13,False)
            GPIO.output(6,False)
            printm("Turn Server Status to Off")

def ckmqttStatus():
    global mqttstate,mqttenabled, mqttfailedat,mqttwasenabled
    cnx= 0
    row = read_control(cnx,'MQTTEnabled')
    mqttenabled = row[2]

    if mqttenabled == 1:
        mqttwasenabled = 1
        row = read_control(cnx,'MQTTStatus')
        mqttstatus = row[2]
        if mqttstatus == 1:
            # enabled and connected Green
            if mqttstate == 0:
                mqttstate = 1
                GPIO.output(26,False)
                GPIO.output(19,True)
                printm("Turn MQTT Status to Green")
        else:
            # Enabled but not connected RED
            GPIO.output(26,True)
            GPIO.output(19,False)
            mqttstate = 0
            mqttfailedat = time.time()
            printm("Turn MQTT Status to Red")
    else:
        # Not enabled, turn off
            if mqttstate == 1 or mqttenabled == 1:
                mqttstate = 0
                mqttwasenabled = 0
                GPIO.output(26,False)
                GPIO.output(19,False)
                printm("Turn MQTT Status to Off")

#
#  Main
#
dstart = datetime.now()
Start =dstart.strftime("%m/%d/%Y %H:%M:%S")

serverstate = -1
serverenabled = 0
serverlastck =time.time()
serverfailedat = time.time()
serverissue = 0

mqttstate = 0
mqttenabled = 0
mqttwasenabled = 0
mqttlastck =time.time()
mqttfailedat = time.time()
mqttissue = 0

intresult = 0
lastlistcount = 0
listlasttime =time.time()
healthevent = 0
event_server = 0
event_mqtt = 0
event_count = 0

while 1:

    #
    dstart = datetime.now()
    LoopTimea =dstart.strftime("%m/%d/%Y %H:%M:%S")
    printm(' Loop Time: ',LoopTimea)
    healthevent = 0
    #
    # Update LED status for Server
    ckServerStatus()
    if time.time() - serverlastck > 120:
        printm("check Server health")
        if serverenabled == 1 and serverstate == 0:
            printm("Server Issue ")
            if serverissue == 0:
                serverissue = 1
            else:
                printm("Long time Server Issue")
                event_server = 2
        else:
            serverissue = 0
            event_server = 0

        serverlastck = time.time()

    #
    # Update LED status for MQTT
    ckmqttStatus()
    if time.time() - mqttlastck > 120:
        printm("check MQTT health")
        if mqttenabled == 1 and mqttstate == 0:
            printm("MQTT Issue ")
            if mqttissue == 0:
                mqttissue = 1
            else:
                printm("Long time MQTT Issue")
                event_mqtt = 2
        else:
            mqttissue = 0
            event_mqtt = 0

        mqttlastck = time.time()

    #
    #
    #printm("check App health")
    led_status = appstate("gw2led.py")
    cmd_status = appstate("gw2cmd.py")
    send_status = appstate("gw2send.py")
    cmd_wr = appstate("gw2wireless.py")
    cmd_hc = appstate("gw2healthcheck.py")

    if led_status == 'Down' or cmd_status == 'Down' or send_status == 'Down' or cmd_wr == 'Down' or cmd_hc == 'Down':
        # allow time for systemd to restart to complete
        printm("Waiting App to Restart")
        time.sleep(2)
        led_status = appstate("gw2led.py")
        cmd_status = appstate("gw2cmd.py")
        send_status = appstate("gw2send.py")
        cmd_wr = appstate("gw2wireless.py")
        cmd_hc = appstate("gw2healthcheck.py")
        if led_status == 'Down' or cmd_status == 'Down' or send_status == 'Down' or cmd_wr == 'Down' or cmd_hc == 'Down':
            healthevent = 1

    #
    # Check Temperatures of CPU/GPU
    srecord = read_control(0,"CPUTemp")
    vrecord = read_control(0,"GPUTemp")
    if float(vrecord[3]) > 160.0 or float(srecord[3]) > 160.0:
        healthevent = 1
        printm("CPU/GPU Temperature issue",srecord[3]," - ",vrecord[3])

    # Verify Ethernet is connected, check carrier value
    networksr = os.popen("netstat -r |grep default").read()
    #printm(networksr)
    networksr = networksr.replace("    ","")
    networksr = networksr.replace("   "," ")
    networksr = networksr.replace("  "," ")
    networksr = networksr.replace("\n","")
    networksr =  networksr.split(' ')
    #printm(networksr)
    networkmax = length_hint(networksr)-1    # handle network read error
    testnetwork =networksr[networkmax]
    try:
        eth0carrier = os.popen(f"cat /sys/class/net/{testnetwork}/carrier ").read()
        #printm("Test for Network Active")
        eth0carrier = int(eth0carrier)
    except:
        ethocarier =0

    if eth0carrier == 0:
        printm("No carrier --> No Network Connnection!")
        healthevent = 1

    printm("Temperature and Network Test Done")
    # future add test for wifi


    # If Counts are the same then issue
    #
    if time.time() - listlasttime > 190:        # Check RX count every 6 minutes
        printm("check RX health")
        sql = "select sum(count) from sensorlist;"
        results = dbsql2(0,sql)
        if results[0] :
            intresult = results[0]
        else:
            intresult = 0

        if lastlistcount == intresult:
            event_count = 1
            listlasttime = time.time()
            printm(f'Count Stopped at {intresult}.')
        else:
            lastlistcount = intresult
            event_count = 0
            listlasttime = time.time()
            #printm("List count is increasing, count is ",intresult)

    #printm("Application Health is ",healthevent)
    if healthevent == 0 and (event_server > 0 or event_mqtt > 0 ):
            healthevent = 2
    elif healthevent == 0 and event_count > 0:
        healthevent = 1

    if healthevent == 1:
        printm(f"Activate Failure Alert S- {event_server}  M- {event_mqtt}  Rx- {event_count}")
        # There is a system problem Relay OFF
        GPIO.output(24,True)
        # turn System LED RED
        GPIO.output(21,True)
        GPIO.output(20,False)

    elif healthevent == 2:
        printm("Activate Failure Alert Relay Only")
        GPIO.output(24,True)   # Relay only Leds show failure

    else:
        # All is well Relay On
        GPIO.output(24,False)
        # Lights to Green
        GPIO.output(21,False)
        GPIO.output(20,True)


    time.sleep(20)

