# mqtt Utilities
import time
from datetime import datetime
import paho.mqtt.client as mqtt
import json
#import MySQLdb
from dbfunctions import dbopen,write_control_n,write_control_t,write_control,read_control,dbsql,dbsqlw,dbopeninit,dbsql2
from utilities import getgwid, set_prefix, printm

import sparkplug_b as sparkplug
from sparkplug_b import *

import sys
sys.path.insert(0, "../../../client_libraries/python/")

class AliasMap:
    Next_Server = 0
    Rebirth = 1
    Reboot = 2
    Dataset = 3
    Node_Metric0 = 4
    Node_Metric1 = 5
    Node_Metric2 = 6
    Node_Metric3 = 7
    Device_Metric0 = 8
    Device_Metric1 = 9
    Device_Metric2 = 10
    Device_Metric3 = 11
    My_Custom_Motor = 12

subscribe = read_control(0, 'MQTTTopic')
subscribe = subscribe[3]
validsub = subscribe.find("/")
if validsub != -1:
     printm("Valid topic: ", subscribe[:validsub])
     myGroupId = subscribe[validsub+1:]
else:
     printm("Invalid topic structure")
     myGroupId = "JemRFDevices"
printm("Group ID:", myGroupId)
nodeId = read_control(0,'NodeId')
nodeId = 'JRF' + nodeId[3]
serverUrl = read_control(0,'MQTTBroker')
serverUrl = serverUrl[3]

def decode_message(json_string):
    global sensor_id
    decoded_json = json.loads(json_string)
    sensors_list = decoded_json["sensors"]
    # Iterate over each sensor object in the list
    for sensor in sensors_list:
        # Print the "sensor" value for each sensor object
        #print("Sensor:", sensor["sensor"], ", Value:", sensor["value"])
        sensor_id = sensor['sensor']
        sensor_value = sensor['value']
        #print(sensor_id[:2],' - ',sensor_id[3:5])
        sensortype = sensor_id[:2]

        if (sensortype == "83"):
                sensor_code = "RSSI"
                device_type = 'SignalStrength'
                device_units = 'dB'

        elif sensortype == "63":
                sensor_code = "TMPA"
                device_type = 'Temperature'
                temflag = read_control(0,'TempScale')
                device_units = 'F'
                if temflag[2] == 0:
                     device_units = 'C'
        elif sensortype == "66":
                sensor_code = "STATE"
                device_type = 'State'
                device_units = 'boolean'
        elif sensortype == "96":
                sensor_code = "BATT"
                device_type = 'Voltage'
                device_units = 'V'
        elif sensortype == "65":
                sensor_code = "HUM"
                device_type = 'Humidity'
                device_units = '%'
        elif sensortype == "57":
                sensor_code = "ANAA"
                device_type = 'LightLevel'
                device_units = 'Lum'
        elif sensortype == "56":
                sensor_code = "ANAB"
                device_type = 'Analog'
                device_units = 'V'

        else:
                sensor_code = "UNKWN"
                device_type = 'Unknown'
                device_units = 'X'

        return [sensor_id,sensor_code,sensor_value,device_type,device_units]

def on_publish(client, userdata, mid, reason_code, properties):
    # reason_code and properties will only be present in MQTTv5. It's always unset in MQTTv3
    try:
        userdata.remove(mid)
    except KeyError:
        print("on_publish() is called with a mid not present in unacked_publish")
        print("This is due to an unavoidable race-condition:")
        print("* publish() return the mid of the message sent.")
        print("* mid from publish() is added to unacked_publish by the main thread")
        print("* on_publish() is called by the loop_start thread")
        print("While unlikely (because on_publish() will be called after a network round-trip),")
        print(" this is a race-condition that COULD happen")
        print("")
        print("The best solution to avoid race-condition is using the msg_info from publish()")
        print("We could also try using a list of acknowledged mid rather than removing from pending list,")
        print("but remember that mid could be re-used !")

def mqtt_establish():
    global unacked_publish,mqttc, nodeId

    unacked_publish = set()
    mqttc = mqtt.Client(mqtt.CallbackAPIVersion.VERSION2)

    write_control_t(0,'MQTTrc','No Broker')
    mqttc.user_data_set(unacked_publish)
    username = read_control(0,'MQTTUser')
    password = read_control(0,'MQTTPass')
    broker = read_control(0,'MQTTBroker')
    port = read_control(0,'MQTTPort')
    username = username[3]
    password = password[3]
    broker = broker[3]
    port = port[3]
    mqttc.on_connect = on_connect
    mqttc.on_message = on_message
    printm(f"Connecting to MQTT broker {broker} on port {port} with username {username} and password {password}")
    mqttc.username_pw_set(username, password)
    mqttc.connect(broker, int(port), 60)
    #printm("MQTT Connected and callbacks enabled")
    mqttc.loop_start()
    printm("MQTT Connected and callbacks enabled")

def mqtt_publish(topic, message):
    global unacked_publish,mqttc
    wait_time = time.time()
    # Our application produce some messages# Wait for all message to be published
    while len(unacked_publish) and (time.time() -wait_time) > 50000:
        time.sleep(0.1)
    #print(time.time() -wait_time )
    msg_info = mqttc.publish(topic,message, 0, False) # qos=1)

    unacked_publish.add(msg_info.mid)
    msg_info.wait_for_publish()

# Due to race-condition described above, the following way to wait for all publish is safer
#msg_info.wait_for_publish()
#msg_info2.wait_for_publish()
def mqtt_exit():
    global unacked_publish,mqttc
    mqttc.disconnect()

    mqttc.loop_stop()

def mqtt_misc():
    global unacked_publish,mqttc
    mqttc.loop_misc()

# The callback for when the client receives a CONNACK response from the server.
def on_connect(client, userdata, flags, reason_code, properties):
    global nodeId, myGroupId
    write_control_t(0,'MQTTrc',str(reason_code))
    if reason_code == "Success":
        printm("Connected with result code"+ str(reason_code))
        write_control_n(0,'MQTTStatus',1)
        write_control_n(0,'MQTTrc',1)
    else:
        printm("Failed to connect with result code-:  "+ str(reason_code))
        write_control_n(0,'MQTTrc',5)
        write_control_n(0,'MQTTStatus',0)
        if reason_code == 1:
             write_control_t(0,'MQTTrc','Incorrect protocol version')
        elif reason_code == 2:
             write_control_t(0,'MQTTrc','Invalid client identifier')
        elif reason_code == 3:
             write_control_t(0,'MQTTrc','Server Unavailable')
        elif reason_code == 4:
             write_control_t(0,'MQTTrc','Bad Username or Password')
        elif reason_code == 5:
             write_control_t(0,'MQTTrc','Not Authorized')
        printm("Error Code Updated")
        '''
        The value of rc indicates success or not:
            0: Connection successful
            1: Connection refused - incorrect protocol version
            2: Connection refused - invalid client identifier
            3: Connection refused - server unavailable
            4: Connection refused - bad username or password
            5: Connection refused - not authorized
            6-255: Currently unused.        '''

        sys.exit()

    # Subscribing in on_connect() means that if we lose the connection and
    # reconnect then subscriptions will be renewed.
    subscribe = read_control(0, 'MQTTTopic')
    subscribe = subscribe[3]
    validsub = subscribe.find("/")
    if validsub != -1:
        printm("Valid topic: ", subscribe[:validsub])
        myGroupId = subscribe[validsub+1:]
    else:
        printm("Blank or Bad topic structure")
        myGroupId = "JemRFDevices"

    mqtt_format = read_control(0,'MQTTFormat')
    mqtt_format = mqtt_format[2]
    if mqtt_format == 4:
        mqttc.subscribe("spBv1.0/" + myGroupId + "/NCMD/" + nodeId + "/#")
        mqttc.subscribe("spBv1.0/" + myGroupId + "/DCMD/" + nodeId + "/#")
        printm("Subscribed to device commands: spBv1.0/" + myGroupId + "/NCMD/" + nodeId + "/#")
    else:
        printm("Subscribed to " + myGroupId)
        mqttc.subscribe(myGroupId)

######################################################################
# The callback for when a PUBLISH message is received from the server.
######################################################################
def on_message(client, userdata, msg):
    global nodeId, myGroupId
    printm("Message arrived: " + msg.topic + " " + str(msg.payload))
    tokens = msg.topic.split("/")
    #printm(tokens)
    printm("--------------------------------")

    if tokens[0] == "spBv1.0" and tokens[1] == myGroupId and (tokens[2] == "NCMD" or tokens[2] == "DCMD") and tokens[3] == nodeId:
        inboundPayload = sparkplug_b_pb2.Payload()
        inboundPayload.ParseFromString(msg.payload)
        for metric in inboundPayload.metrics:
            #printm("Decoded Metric: " + metric.name)
            if metric.name == "Node Control/Next Server":
                # 'Node Control/Next Server' is an NCMD used to tell the device/client application to
                # disconnect from the current MQTT server and connect to the next MQTT server in the
                # list of available servers.  This is used for clients that have a pool of MQTT servers
                # to connect to.
                printm ("'Node Control/Next Server' is not implemented in this example.")
            elif metric.name == "Node Control/Rebirth" :
                printm ("'Node Control/Rebirth")
                # 'Node Control/Rebirth' is an NCMD used to tell the device/client application to resend
                # its full NBIRTH and DBIRTH again.  MQTT Engine will send this NCMD to a device/client
                # application if it receives an NDATA or DDATA with a metric that was not published in the
                # original NBIRTH or DBIRTH.  This is why the application must send all known metrics in
                # its original NBIRTH and DBIRTH messages.
                publishBirth()
            elif metric.name == "Node Control/Reboot":
                printm ("'Node Control/Reboot")
                # 'Node Control/Reboot' is an NCMD used to tell a device/client application to reboot
                # This can be used for devices that need a full application reset via a soft reboot.
                # In this case, we fake a full reboot with a republishing of the NBIRTH and DBIRTH
                # messages.
                publishBirth()
            else:
                printm ("Unknown command: " + metric.name)
    else:
        printm ("On Message: Unknown command...")

    printm("Done publishing")
######################################################################
def publishBirth():
    global sensor_id
    printm("Rebirth message")
    publishNodeBirth(nodeId)
    publishDeviceBirth(nodeId,sensor_id[3:5])
    write_control_n(0,'MQTTStatus',0)

def on_disconnect(mqttc, userdata, rc):

    printm("disconnecting reason  "+str(rc))
    mqttc.connect_flag=False
    mqttc.disconnect_flag=True

def enablesparkplug(nodeId):
    global myGroupId, mqttc
    printm("Sending birth message for node: " + nodeId)
    #spBv1.0/JemRFDevices/NBIRTH/JRF03516124
    #{"timestamp":"1711930675","metrics":[{"name":"JRF03516124","timestamp":"1711930675","dataType":"String","value":"Gateway"}],"seq":0}
    #
    topic = "spBv1.0/"+ myGroupId + "/" + "NBIRTH" + "/"+nodeId
    #print("Publishing birth message to topic:", topic)
    payload = '{"timestamp":"' + str(int(time.time())) + '","metrics":[{"name":"' + nodeId + '","timestamp":"' + str(int(time.time())) + '","dataType" : "12","value":"Gateway"}],"seq":0}'
    #print("Payload:", payload)
    #mqtt_publish(topic, payload)
    publishNodeDeath(nodeId)
    publishNodeBirth(nodeId)

def publishNodeBirth(myNodeName):
    global myGroupId, mqttc
    printm ("Publishing Node Birth")

    # Create the node birth payload
    payload = sparkplug.getNodeBirthPayload()

    # Set up the Node Controls

    # Add some regular node metrics
    addMetric(payload, "Gateway", AliasMap.Node_Metric0, MetricDataType.String, myNodeName)

    # Create a DataSet (012 - 345) two rows with Int8, Int16, and Int32 contents and headers Int8s, Int16s, Int32s and add it to the payload
    # Add a metric with a custom property
    # Publish the node birth certificate
    byteArray = bytearray(payload.SerializeToString())
    #print("topic: spBv1.0/" + myGroupId + "/NBIRTH/" + myNodeName )
    mqttc.publish("spBv1.0/" + myGroupId + "/NBIRTH/" + myNodeName, bytes(byteArray), 0, False)

######################################################################
# Publish the DBIRTH certificate
######################################################################
def publishDeviceBirth(myNodeName, myDeviceName):
    global myGroupId, mqttc
    printm("Publishing Device Birth for node: " + myNodeName + " device: " + myDeviceName)

    # Get the payload
    payload = sparkplug.getDeviceBirthPayload()

    # Add some device metrics
    addMetric(payload, "Voltage", None, MetricDataType.Float, 0)    # No alias in UDT members
    addMetric(payload, "SignalStrength", None, MetricDataType.Float, 0)    # No alias in UDT members
    #
    addMetric(payload, "Temperature", None, MetricDataType.Float, 0)    # No alias in UDT members
    addMetric(payload, "Humidity", None, MetricDataType.Float, 0)    # No alias in UDT members
    #addMetric(payload, "Lightlevel", None, MetricDataType.Float, 0)    # No alias in UDT members

    # Publish the initial data with the Device BIRTH certificate
    totalByteArray = bytearray(payload.SerializeToString())
    mqttc.publish("spBv1.0/" + myGroupId + "/DBIRTH/" + myNodeName + "/" + myDeviceName, bytes(totalByteArray), 0, False)
   #client.connect(serverUrl, 1883, 60)
    #printm("Published initial data")
######################################################################

def publishNodeDeath(nodeId):
    global mqttc,serverUrl,myGroupId
    printm ("Publishing Node Death",nodeId)
    deathPayload = sparkplug.getNodeDeathPayload()

    # Start of main program - Set up the MQTT client connection
    deathByteArray = bytearray(deathPayload.SerializeToString())
    mqttc.will_set("spBv1.0/"+ myGroupId + "/NDEATH/" + nodeId, deathByteArray, 0, False)

    mqttc.connect(serverUrl, 1883, 60)
    # Short delay to allow connect callback to occur
    time.sleep(.1)

def spsenddata(sensor_id,sensor_type,sensor_unit,sensor_value,rxtime):
    global mqttc, nodeId, myGroupId
    payload = sparkplug.getDdataPayload()

    #printm("sp send id=",sensor_id," -type= ",sensor_type," -unit= ",sensor_unit," -value= ",sensor_value," - ",rxtime)

    #printm("sp payload: ",payload.timestamp)
    dt_object = datetime.strptime(rxtime, '%Y/%m/%d %H:%M:%S')
    #printm("dt_object = ",dt_object)
    rxtimestamp = time.mktime(dt_object.timetuple())
    #printm("rxpayload: ",int(round(rxtimestamp * 1000)))
    payload.timestamp = int(round(rxtimestamp * 1000))

    #addMetric(payload, "Timestamp", None, MetricDataType.String, str(int(time.time())))

    #print("SP Sending data message for sensor:", sensor_id)
    # Add some random data to the inputs   Reading is Temp,Hum, Volt, RSSI
    if sensor_unit == "datetime":
        addMetricd(payload, sensor_type, None, MetricDataType.String, sensor_value, rxtimestamp )
    elif sensor_unit == "string":
        addMetricd(payload, sensor_type, None, MetricDataType.String, sensor_value, rxtimestamp )
    else:
        addMetricd(payload, sensor_type , None, MetricDataType.Float, float(sensor_value ), rxtimestamp )
    #print(payload)
    # Publish a message data
    byteArray = bytearray(payload.SerializeToString())
    #print("Sp-Topic: spBv1.0/" + myGroupId + "/DDATA/" + nodeId + "/" + sensor_id)
    (rc, mid) = mqttc.publish("spBv1.0/" + myGroupId + "/DDATA/" + nodeId + "/" + sensor_id, bytes(byteArray), 0, False)
    #print("SP Data published")
    #print("SP Data published",rc)
    if rc == mqtt.MQTT_ERR_SUCCESS:
        #printm("Publish success!")
        nullvalue = 0
    else:
        write_control_n(0,'MQTTrc',5)
        printm("Publish failed with error code: " + str(rc))
        write_control_t(0,'MQTTrc',str(rc))
        write_control_n(0,'MQTTStatus',0)
        printm("Attempt reconnection")
        mqtt_exit()
        mqtt_establish()